from satella.coding.typing import ExceptionList
import typing as tp


class expect_exception:
    """
    A context manager to use as following:

    >>> a = {'test': 2}
    >>> with expect_exception(KeyError, ValueError, 'KeyError not raised'):
    >>>     a['test2']

    If other exception than the expected is raised, it is passed through

    :param exc_to_except: a list of exceptions or a single exception to expect
    :param else_raise: raise a particular exception if no exception is raised.
        This should be a callable that accepts provided args and kwargs and returns
        an exception instance.
    :param args: args to provide to constructor
    :param kwargs: kwargs to provide to constructor
    """
    __slots__ = 'exc_to_except', 'else_raise', 'else_raise_args', 'else_raise_kwargs'

    def __init__(self, exc_to_except: ExceptionList, else_raise: tp.Type[Exception],
                 *args, **kwargs):
        self.exc_to_except = exc_to_except
        self.else_raise = else_raise
        self.else_raise_args = args
        self.else_raise_kwargs = kwargs

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            raise self.else_raise(*self.else_raise_args,
                                  **self.else_raise_kwargs)
        elif not isinstance(exc_val, self.exc_to_except):
            return False
        return True



import unittest

from satella.coding import source_to_function
from satella.coding.predicates import x, build_structure


class TestPredicates(unittest.TestCase):

    def test_source_to_function(self):
        y = source_to_function(x)
        self.assertEqual(y(5), 5)

    def test_build_structure(self):
        a = {x: x*2}
        b = [x, x*2, x*3]
        c = (x, x*2, x*3)
        self.assertEqual(build_structure(a, 2), {2: 4})
        self.assertEqual(build_structure(b, 2), [2, 4, 6])
        self.assertEqual(build_structure(c, 2), (2, 4, 6))
        self.assertEqual(build_structure([(x, [x, x+1, x+2]),
                                          (x+1, [x, x+1, x+2])], 2, dict),
                         {2: [2, 3, 4],
                          3: [2, 3, 4]})
        self.assertEqual(build_structure(2, 2), 2)

    def test_is_instance(self):
        self.assertEqual(x.type()(2), int)
        self.assertTrue(x.is_instance(str, int)('str'))
        self.assertTrue(x.is_instance(str, int)(5))
        self.assertFalse(x.is_instance(str, int)(5.0))

    def test_is_valid_schema(self):
        a = {"test": 5, "test2": "string"}
        self.assertTrue(x.is_valid_schema({"test": "int", "test2": "str"})(a))
        self.assertFalse(x.is_valid_schema({"test3": "str", "test2": "str"})(a))

    def test_has_p(self):
        class Object:
            def __init__(self):
                self.a = [1, 2, 3, 4, 5, 6]
        a = Object()
        self.assertTrue(x.a.has_p(x % 2 == 0)(a))
        a.a = [3, 5, 7, 9]
        self.assertFalse(x.a.has_p(x % 2 == 0)(a))

    def test_int(self):
        p = x.int()
        self.assertEqual(p('2'), 2)

    def test_bool_str(self):
        p = x.int().str()
        self.assertEqual(p(True), '1')
        p = x.false()
        self.assertTrue(p(False))
        self.assertFalse(p(True))

    def test_upper(self):
        p = x.upper()
        self.assertEqual(p('ala ma kota'), 'ALA MA KOTA')

    def test_instanceof(self):
        p = x.instanceof(int)
        self.assertTrue(p(2))
        self.assertFalse(p('2'))

    def test_has_keys(self):
        a = {'hello': 'world', 'hello2': 'world'}
        p = x.has_keys('hello', 'hello2')
        self.assertTrue(p(a))
        del a['hello']
        self.assertFalse(p(a))

    def test_joined_predicates(self):
        p = (x > 2) & (x < 6)
        self.assertTrue(p(4))
        self.assertFalse(p(1))
        self.assertFalse(p(8))

        p = (x < 2) | (x > 6)
        self.assertTrue(p(1))
        self.assertTrue(p(8))
        self.assertFalse(p(4))

    def test_ops(self):
        p = (x + 2) == 2
        self.assertTrue(p(0))
        self.assertFalse(p(1))
        p = (x - 2) == 0
        self.assertTrue(p(2))
        self.assertFalse(p(1))
        p = (x * 2) == 2
        self.assertTrue(p(1))
        self.assertFalse(p(2))
        p = (x / 2) == 1
        self.assertTrue(p(2))
        self.assertFalse(p(1))
        p = (x + 2) % 3 == 0
        self.assertTrue(p(1))
        self.assertFalse(p(2))

    def test_one_of(self):
        p = x.one_of(2, 3)
        self.assertTrue(p(2))
        self.assertTrue(p(3))
        self.assertFalse(p(4))

    def test_contains(self):
        p = x.inside(2)
        self.assertTrue(p([1, 2]))
        self.assertFalse(p([1, 3]))

    def test_getattr(self):
        class A:
            def __init__(self, a=2):
                self.attr = a

        p = x.attr == 2
        self.assertTrue(p(A()))
        self.assertFalse(p(A(3)))

    def test_getitem(self):
        p = x[0] == 1
        self.assertTrue(p([1, 2]))
        self.assertFalse(p([2, 2]))

    def test_len(self):
        p = x.length() == 2
        self.assertTrue(p([1, 2]))
        self.assertFalse(p([]))

    def test_equals(self):
        p = x == 2
        self.assertTrue(p(2))
        self.assertFalse(p(3))
        p = x > 2
        self.assertTrue(p(3))
        self.assertFalse(p(2))
        p = x < 2
        self.assertTrue(p(1))
        self.assertFalse(p(2))
        p = x >= 2
        self.assertTrue(p(2))
        p = x <= 2
        self.assertTrue(p(1))
        self.assertTrue(p(2))
        self.assertFalse(p(3))
        p = x != 2
        self.assertTrue(p(1))
        self.assertFalse(p(2))



import collections
import enum
import random
import typing as tp

from satella.coding.decorators import for_argument
from .jsonify import jsonify
from .merger import merge_series
from .merge_list import merge_list
from .percentile import percentile
from .base64 import b64encode
from .interpol import linear_interpolate
from .words import hashables_to_int
from .predicates import is_subset

__all__ = ['stringify', 'split_shuffle_and_join', 'one_tuple', 'none_if_false',
           'merge_series', 'pad_to_multiple_of_length', 'clip', 'hashables_to_int',
           'jsonify', 'intify', 'percentile', 'b64encode', 'linear_interpolate',
           'merge_list', 'is_subset']

from satella.coding.typing import T, NoArgCallable, Appendable, Number, Predicate


def none_if_false(y: tp.Any) -> tp.Optional[tp.Any]:
    """
    Return None if y is false, else return y

    :param y: value to check
    :return: None if y is false, else y
    """
    if not y:
        return None
    return y


def clip(v: Number, minimum: Number, maximum: Number) -> Number:
    """
    Clip v so it conforms to minimum <= v <= maximum

    :param v: value to clip
    :param minimum: minimum
    :param maximum: maximum
    :return: clipped value
    """
    if v < minimum:
        return minimum
    if v > maximum:
        return maximum
    return v


@for_argument(list)
def pad_to_multiple_of_length(seq: Appendable[T], multiple_of: int,
                              pad_with: tp.Optional[T] = None,
                              pad_with_factory: tp.Optional[NoArgCallable[T]] = None) -> \
        Appendable[T]:
    """
    Make sequence multiple of length, ie. append elements to the sequence
    until it's length is a multiple of multiple_of.

    :param seq: sequence to lengthify
    :param multiple_of: sequence returned will be a multiple of this length.
    :param pad_with: argument with which to pad the sequence
    :param pad_with_factory: a callable/0 that returns an element with which to pad the sequence
    :return: a list with elements
    """
    if pad_with is not None and pad_with_factory is not None:
        raise ValueError('You need to give either pad_with or pad_with_factory')

    if pad_with_factory is None:
        def pad_with_factory():
            return pad_with

    while len(seq) % multiple_of:
        seq.append(pad_with_factory())
    return seq


def _stringify_none(str_none: tp.Optional[str],
                    stringifier: tp.Callable[[tp.Any], str]) -> tp.Optional[str]:
    if str_none:
        return stringifier(None)
    return None


def one_tuple(x: tp.Iterable[T]) -> tp.Iterator[tp.Tuple[T]]:
    """
    Change a sequence of iterables into a sequence that displays each element as
    a part of one-element tuple. Essentially syntactic sugar for:

    >>> for z in x:
    >>>     yield z,

    :param x: sequence to tupleify
    :return: a iterator of one-element tuples
    """
    for z in x:
        yield z,


def intify(v: tp.Any) -> int:
    """
    Attempt to convert v to an int.

    None will be converted to 0.

    Any object will have int() called on it.

    Failing that, it's length will be taken.

    Failing that, ValueError will be raised

    :param v: parameter
    :return: int representation
    """
    if v is None:
        return 0

    try:
        return int(v)
    except (TypeError, ValueError):
        try:
            return len(v)
        except (AttributeError, TypeError, ValueError):
            raise ValueError('Unable to convert %s to int' % (v,))


def split_shuffle_and_join(entries: tp.List[T],
                           whether_to_shuffle: Predicate[T] = lambda x: True,
                           not_shuffled_to_front: bool = True) -> tp.List[T]:
    """
    Split elements in entries into two groups - one group, called True, is the one for which
    whether_to_shuffle(elem) is True, the other is False.

    Shuffle the group True.

    If not_shuffled_to_front, elements in the group False will go at the beginning of the returned
    list, after which will go elements shuffled. If it's False, the not-shuffled elements will be
    at the back of the list.

    Order of the not shuffled elements will be preserved.

    :param entries: list of elements
    :param whether_to_shuffle: a decider to which group does given element belong?
    :param not_shuffled_to_front: if True then not shuffled elements will be put before shuffled,
        else the not shuffled elements will be at the back of the list
    :return: list altered to specification
    """

    not_shuffled, shuffled = [], []
    for elem in entries:
        (shuffled if whether_to_shuffle(elem) else not_shuffled).append(elem)

    random.shuffle(shuffled)

    if not_shuffled_to_front:
        return not_shuffled + shuffled
    else:
        return shuffled + not_shuffled


def stringify(obj: tp.Union[tp.Any], stringifier: tp.Callable[[tp.Any], str] = str,
              recursively: bool = False,
              str_none: bool = False) -> tp.Union[tp.List[str], tp.Dict[str, str], str]:
    """
    Stringify all object:

    ie. if a dict, put every item and key (if a dict is given) through stringify.
        if a list, put every item through stringify
        else just call stringify on it.

    Note that if you use recursively, then dicts and lists are allowed to be valid elements of the
    returned representation!

    Note that enums will be converted to their labels. eg:

    >>> class Enum(enum.Enum):
    >>>     A = 0
    >>> assert stringify(Enum.A) == 'A'

    :param obj: a list or a dict
    :param stringifier: function that accepts any arguments and returns a string representation
    :param recursively: whether to recursively stringify elements, ie. stringify will be called on
        all the children
    :param str_none: whether to return None if given a None. If True, "None" will be returned
        instead
    :return: stringified object
    """
    if isinstance(obj, str):
        y = obj
    elif isinstance(obj, enum.Enum):
        y = obj.name
    elif isinstance(obj, collections.abc.Mapping):
        make_str = (lambda obj2: stringify(obj2, stringifier, True, str_none)) if recursively else \
            stringifier
        y = {make_str(k): make_str(v) for k, v in obj.items()}
    elif isinstance(obj, collections.abc.Sequence):
        make_str = (lambda obj2: stringify(obj2, stringifier, True, str_none)) if recursively else \
            stringifier
        y = [make_str(v) for v in obj]
    elif obj is None:
        y = _stringify_none(str_none, stringifier)
    else:
        y = stringifier(obj)
    return y



import logging
import time
import unittest

import requests

from satella.instrumentation.metrics import MetricData, MetricDataCollection
from satella.instrumentation.metrics import getMetric
from satella.instrumentation.metrics.exporters import metric_data_collection_to_prometheus, \
    PrometheusHTTPExporterThread

logger = logging.getLogger(__name__)


class TestExporters(unittest.TestCase):
    def test_prometheus_timestamps(self):
        a = MetricDataCollection(MetricData('root', 5, timestamp=10))
        self.assertEqual("""root 5 10000\n""", metric_data_collection_to_prometheus(a))

    def test_internal_metrics(self):
        metric = getMetric('internal_metric', 'int', internal=True)
        metric.runtime(2)
        self.assertNotIn('internal_metric', metric_data_collection_to_prometheus(
            getMetric('').to_metric_data()))

    def test_prometheus(self):
        a = MetricDataCollection([MetricData('root.metric', 3, {'k': 2, 'm': '"'}),
                                  MetricData('root.metric', 6, {'k': 4})])
        b = metric_data_collection_to_prometheus(a)
        self.assertIn("""root_metric{k="4"} 6""", b)

    def test_exporter_http_server(self):
        with PrometheusHTTPExporterThread('localhost', 1025):
            metr = getMetric('test.metric', 'int')
            metr.runtime(5)
            time.sleep(0.5)
            data = requests.get('http://localhost:1025/metrics')
            self.assertEqual(200, data.status_code)
            self.assertIn('test_metric 5', data.text)
            data2 = requests.get('http://localhost:1025/404')
            self.assertEqual(404, data2.status_code)



import copy
import typing as tp

from satella.coding.recast_exceptions import rethrow_as
from satella.exceptions import ConfigurationValidationError
from .base import Descriptor, must_be_type, ConfigDictValue
from .registry import register_custom_descriptor


@register_custom_descriptor('list', is_plain=False)
class List(Descriptor):
    """
    This must be a list, made of entries of a descriptor (this is optional)
    """
    __slots__ = ('type_descriptor',)

    CHECKERS = [must_be_type(list, tuple)]
    BASIC_MAKER = list

    def __init__(self, type_descriptor: tp.Optional[Descriptor] = None):
        super().__init__()
        self.type_descriptor = type_descriptor or Descriptor()  # type: Descriptor

    def __call__(self, value: ConfigDictValue) -> tp.List:
        value = super().__call__(value)

        return [self.type_descriptor(p) for p in value]

    def __str__(self):
        return 'List(%s)' % (self.type_descriptor,)


DictDescriptorKey = tp.NewType('DictDescriptorKey', Descriptor)


def create_key(descriptor: Descriptor, name: str, optional: bool = False,
               default: tp.Optional[tp.Any] = None) -> DictDescriptorKey:
    descriptor.name = name
    descriptor.optional = optional
    descriptor.default = default

    return descriptor


@register_custom_descriptor('caster', is_plain=False)
class Caster(Descriptor):
    """
    A value must be ran through a function.

    Use like:

    >>> class Environment(enum.IntEnum):
    >>>     PRODUCTION = 0
    >>> assert Caster(Environment)(0) == Environment.PRODUCTION
    """

    def __init__(self, to_cast: tp.Callable[[tp.Any], tp.Any]):
        self.to_cast = to_cast

    @rethrow_as(ValueError, ConfigurationValidationError)
    def __call__(self, value: ConfigDictValue):
        return self.to_cast(value)


@register_custom_descriptor('dict', is_plain=False)
class Dict(Descriptor):
    """
    This entry must be a dict, having at least specified keys.

    Use like:

    >>> Dict([
    >>>     create_key(String(), 'key_s'),
    >>>     create_key(Integer(), 'key_i'),
    >>>     create_key(Float(), 'key_f'),
    >>>     create_key(String(), 'key_not_present', optional=True,
    >>>                default='hello world'),
    >>>     create_key(IPv4(), 'ip_addr')
    >>>])
    """
    __slots__ = ('keys', 'unknown_key_mapper')

    BASIC_MAKER = dict
    CHECKERS = [must_be_type(dict)]

    UnknownKeyHandlerType = tp.Callable[[str, ConfigDictValue], tp.Any]

    def __init__(self, keys: tp.List[DictDescriptorKey],
                 unknown_key_mapper: UnknownKeyHandlerType = lambda str,
                                                                    data: data):
        super().__init__()
        self.keys = {item.name: item for item in keys}  # type: tp.Dict[str, DictDescriptorKey]
        self.unknown_key_mapper = unknown_key_mapper  # Dict.UnknownKeyHandlerType

    def __str__(self):
        return 'Dict(%s)' % (self.keys,)

    def __call__(self, value: ConfigDictValue) -> dict:
        value = copy.copy(value)
        if not isinstance(value, dict):
            raise ConfigurationValidationError('value passed was not a dict: %s' % (value,))
        value = super().__call__(value)

        if not isinstance(value, dict):
            raise ConfigurationValidationError('value received from descriptor was not a dict: %s'
                                               % (value,))

        output = {}

        for key, key_descriptor in self.keys.items():
            try:
                v = value.pop(key)
            except KeyError:
                if key_descriptor.optional:
                    output[key] = key_descriptor.default
                else:
                    raise ConfigurationValidationError('required key %s not found' % (key,))
            else:
                output[key] = key_descriptor(v)

        for k, v in value.items():
            output[k] = self.unknown_key_mapper(k, v)

        return output


@register_custom_descriptor('union', is_plain=False)
class Union(Descriptor):
    """
    The type of one of the child descriptors. If posed as such:

    Union(List(), Dict())

    then value can be either a list or a dict
    """
    __slots__ = ('descriptors',)

    def __init__(self, *descriptors: tp.List[Descriptor]):
        super().__init__()
        self.descriptors = descriptors

    def __call__(self, value: tp.Any) -> tp.Any:
        for descriptor in self.descriptors:
            try:
                return descriptor(value)
            except ConfigurationValidationError:
                continue
        raise ConfigurationValidationError('Could not match value %s to a descriptor' % (value,))



import typing as tp
from concurrent.futures import Future
from threading import Thread

from satella.coding.decorators.decorators import wraps
from satella.coding.sequences.sequences import infinite_iterator
from satella.coding.typing import T


def run_as_future(fun):
    """
    A decorator that accepts a function that should be executed in a separate thread,
    and a Future returned instead of it's result, that will enable to watch the function for
    completion.

    The created thread will be non-demonic

    Example usage:

    >>> @run_as_future
    >>> def parse_a_file(x: str):
    >>>     ...
    >>> fut = parse_a_file('test.txt')
    >>> result = fut.result()
    """

    @wraps(fun)
    def inner(*args, **kwargs):
        fut = Future()
        fut.set_running_or_notify_cancel()

        def separate_target():
            try:
                fut.set_result(fun(*args, **kwargs))
            except Exception as e:
                fut.set_exception(e)

        Thread(target=separate_target).start()
        return fut

    return inner


def parallel_execute(callable_: tp.Callable[[T], Future],
                     args: tp.Iterable[T],
                     kwargs: tp.Iterable[dict] = infinite_iterator(return_factory=dict)):
    """
    Execute a number of calls to callable in parallel.

    Callable must be a function that accepts arguments and returns a plain Python future.

    Return will be an iterator that will yield every value of the iterator,
    or return an instance of exception, if any of the calls excepted.

    :param callable_: a callable that returns futures
    :param args: an iterable of arguments to provide to the callable
    :param kwargs: an iterable of keyword arguments to provide to the callable
    :return: an iterator yielding every value (or exception instance if thew) of the future
    """
    futures = [callable_(*arg, **kwarg) for arg, kwarg in zip(args, kwargs)]
    for future in futures:
        try:
            yield future.result()
        except Exception as e:
            yield e



import logging
import math
import typing as tp

from satella.coding.decorators.decorators import wraps
from satella.coding.recast_exceptions import rethrow_as
from satella.coding.typing import T

logger = logging.getLogger(__name__)

_SETTABLE_KEYS = {'_Proxy__obj', '_Proxy__wrap_operations'}
Number = tp.Union[float, int]


def wrap_operation(fun):
    @wraps(fun)
    def inner(self: 'Proxy', *args, **kwargs):
        result = fun(self, *args, **kwargs)
        if self._Proxy__wrap_operations:
            result = self.__class__(result)
        return result

    return inner


class Proxy(tp.Generic[T]):
    """
    A base class for classes that try to emulate some other object.

    They will intercept all calls and place them on the target object.

    Note that in-place operations will return the Proxy itself, whereas simple addition will shed
    this proxy, returning object wrapped plus something.

    Only __isinstance__ check is not overridden, due to Python limitations.

    Please access this object in your descendant classes via

    >>> getattr(self, '_Proxy__obj')

    Please note that this class does not overload the descriptor protocol,
    not the pickle interface!

    Note that this overloads __repr__, __str__ and __dir__, which may prove confusing.
    Handle this in your descendant classes.

    If wrap_operations is set, the following code will be executed on the result:

    >>> a = a.__add__(b)
    >>> return self.__class__(a)

    Wrapped operations include all arithmetic operations and bitwise operations,
    except for divmod, but including concat, rounding, truncing and ceiling.

    :param object_to_wrap: object to wrap
    :param wrap_operations: whether results of operations returning something else should be
        also proxied.
    """

    def __init__(self, object_to_wrap: T, wrap_operations: bool = False):
        self.__obj = object_to_wrap  # type: T
        self.__wrap_operations = wrap_operations  # type: bool

    def __call__(self, *args, **kwargs):
        return self.__obj(*args, **kwargs)

    def __getitem__(self, item):
        return self.__obj[item]

    def __setitem__(self, key, value) -> None:
        self.__obj[key] = value

    def __delitem__(self, key) -> None:
        del self.__obj[key]

    def __setattr__(self, key, value) -> None:
        if key in _SETTABLE_KEYS:
            super().__setattr__(key, value)
        else:
            setattr(self.__obj, key, value)

    def __getattr__(self, item):
        return getattr(self.__obj, item)

    def __delattr__(self, item) -> None:
        delattr(self.__obj, item)

    def __int__(self) -> int:
        return int(self.__obj)

    @rethrow_as(AttributeError, TypeError)
    def __float__(self) -> float:
        return float(self.__obj)

    @rethrow_as(AttributeError, TypeError)
    def __complex__(self) -> complex:
        return complex(self.__obj)

    def __str__(self) -> str:
        return str(self.__obj)

    @wrap_operation
    def __add__(self, other):
        return self.__obj + other

    def __iadd__(self, other) -> 'Proxy':
        self.__obj += other
        return self

    @wrap_operation
    def __sub__(self, other):
        return self.__obj - other

    def __isub__(self, other) -> 'Proxy':
        self.__obj -= other
        return self

    @wrap_operation
    def __mul__(self, other):
        return self.__obj * other

    def __divmod__(self, other) -> tp.Tuple[Number, Number]:
        return divmod(self.__obj, other)

    @wrap_operation
    def __floordiv__(self, other):
        return self.__obj // other

    def __rdivmod__(self, other):
        return divmod(other, self.__obj)

    @wrap_operation
    def __truediv__(self, other):
        return self.__obj / other

    def __imul__(self, other) -> 'Proxy':
        self.__obj *= other
        return self

    def __itruediv__(self, other) -> 'Proxy':
        self.__obj /= other
        return self

    def __ifloordiv__(self, other) -> 'Proxy':
        self.__obj //= other
        return self

    @wrap_operation
    def __rshift__(self, other):
        return self.__obj >> other

    @wrap_operation
    def __lshift__(self, other):
        return self.__obj << other

    def __ilshift__(self, other) -> 'Proxy':
        self.__obj <<= other
        return self

    def __irshift__(self, other) -> 'Proxy':
        self.__obj >>= other
        return self

    def __iter__(self) -> tp.Iterator:
        return iter(self.__obj)

    def __len__(self) -> int:
        return len(self.__obj)

    def __contains__(self, item) -> bool:
        return item in self.__obj

    def __hash__(self) -> int:
        return hash(self.__obj)

    def __eq__(self, other) -> bool:
        return self.__obj == other

    def __ne__(self, other) -> bool:
        return self.__obj != other

    def __or__(self, other):
        return self.__obj | other

    def __and__(self, other):
        return self.__obj & other

    def __le__(self, other) -> bool:
        return self.__obj <= other

    def __lt__(self, other) -> bool:
        return self.__obj < other

    def __ge__(self, other) -> bool:
        return self.__obj >= other

    def __gt__(self, other) -> bool:
        return self.__obj > other

    def __enter__(self):
        return self.__obj.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.__obj.__exit__(exc_type, exc_val, exc_tb)

    def __repr__(self) -> str:
        return repr(self.__obj)

    def __oct__(self) -> str:
        return oct(self.__obj)

    def __hex__(self) -> str:
        return hex(self.__obj)

    @wrap_operation
    def __abs__(self):
        return abs(self.__obj)

    def __bool__(self) -> bool:
        return bool(self.__obj)

    def __format__(self, format_spec):
        return self.__obj.__format__(format_spec)

    def __next__(self):
        return next(self.__obj)

    @wrap_operation
    def __xor__(self, other):
        return self.__obj ^ other

    @wrap_operation
    def __radd__(self, other):
        return other + self.__obj

    @wrap_operation
    def __rsub__(self, other):
        return other - self.__obj

    @wrap_operation
    def __rmul__(self, other):
        return other * self.__obj

    @wrap_operation
    def __rtruediv__(self, other):
        return other / self.__obj

    @wrap_operation
    def __rfloordiv__(self, other):
        return other // self.__obj

    @wrap_operation
    def __rlshift__(self, other):
        return other << self.__obj

    @wrap_operation
    def __rrshift__(self, other):
        return other >> self.__obj

    @wrap_operation
    def __rpow__(self, other):
        return other ** self.__obj

    @wrap_operation
    def __ror__(self, other):
        return other | self.__obj

    @wrap_operation
    def __rand__(self, other):
        return other & self.__obj

    @wrap_operation
    def __rxor__(self, other):
        return other ^ self.__obj

    def __ior__(self, other) -> 'Proxy':
        self.__obj |= other
        return self

    def __iand__(self, other) -> 'Proxy':
        self.__obj &= other
        return self

    def __ixor__(self, other) -> 'Proxy':
        self.__obj ^= other
        return self

    def __await__(self):
        return self.__obj.__await()

    def __aenter__(self):
        return self.__obj.__aenter__()

    def __aexit__(self, exc_type, exc_val, exc_tb):
        return self.__obj.__aexit__(exc_type, exc_val, exc_tb)

    def __aiter__(self):
        return self.__obj.__aiter__()

    def __reversed__(self):
        return reversed(self.__obj)

    @wrap_operation
    def __pow__(self, power, modulo=None):
        return pow(self.__obj, power, modulo)

    def __ipow__(self, other) -> 'Proxy':
        self.__obj **= other
        return self

    @wrap_operation
    def __neg__(self):
        return -self.__obj

    @wrap_operation
    def __pos__(self):
        return +self.__obj

    @wrap_operation
    def __invert__(self):
        return ~self.__obj

    @wrap_operation
    def __matmul__(self, other):
        return self.__obj @ other

    def __imatmul__(self, other):
        self.__obj @= other
        return self

    @wrap_operation
    def __rmatmul__(self, other):
        return other @ self.__obj

    def __index__(self) -> int:
        return self.__obj.__index__()

    @wrap_operation
    @rethrow_as(AttributeError, TypeError)
    def __round__(self, n: int = 0):
        return round(self.__obj, n)

    @wrap_operation
    @rethrow_as(AttributeError, TypeError)
    def __trunc__(self):
        if hasattr(self.__obj, '__trunc__'):
            result = self.__obj.__trunc__()
        else:
            result = math.trunc(self.__obj)
        return result

    @wrap_operation
    @rethrow_as(AttributeError, TypeError)
    def __floor__(self):
        if hasattr(self.__obj, '__floor__'):
            result = self.__obj.__floor__()
        else:
            result = math.floor(self.__obj)

        return result

    @wrap_operation
    @rethrow_as(AttributeError, TypeError)
    def __ceil__(self):
        if hasattr(self.__obj, '__ceil__'):
            result = self.__obj.__ceil__()
        else:
            result = math.ceil(self.__obj)
        return result

    def __dir__(self) -> tp.Iterable[str]:
        return dir(self.__obj)

    @wrap_operation
    @rethrow_as(AttributeError, TypeError)
    def __concat__(self, other):
        return self.__obj.__concat__(other)

    @rethrow_as(AttributeError, TypeError)
    def __iconcat__(self, other) -> 'Proxy':
        self.__obj.__iconcat__(other)
        return self

    @rethrow_as(AttributeError, TypeError)
    def __length_hint__(self):
        return self.__obj.__length_hint__()



import base64


def b64encode(content: bytes) -> str:
    """
    Syntactic sugar for:

    >>> import base64
    >>> y = base64.b64encode(content).decode('utf-8')

    Since `base64.b64decode(str)` returns bytes, the reverse is not provided.

    :param content: content to encode
    :return: content encoded as a string
    """
    return base64.b64encode(content).decode('utf-8')



import collections
import typing as tp

from satella.coding.recast_exceptions import silence_excs
from satella.coding.typing import T

KeyArg = tp.Tuple[tp.Union[int, slice], tp.Union[int, slice]]


def _cleanup_key(inst: tp.Union[int, slice], max_count: int):
    if isinstance(inst, slice):
        if not (inst.start is None and inst.stop is None and inst.step is None):
            raise IndexError('Custom slicing is not supported!')
        return Ellipsis
    elif isinstance(inst, int):
        if isinstance(inst, int):
            if inst < 0:
                inst += max_count
        return inst


class SparseMatrix(tp.Generic[T]):
    """
    A matrix of infinite size, that supports assignments.

    Set elements like this:

    >>> sm[1, 2] = 5
    >>> sm[:,1] = [5]
    >>> sm[1,:] = [5]
    >>> sm[:,:] = [[5]]

    where the first argument is the number of the column, counted from 0,
    and the second one is the number of the row, also counted from 0

    Note that custom slicing (ie. slices which are not :) will not be supported.
    Negative indices are supported.

    Undefined elements are considered to be of value None.

    Iterating over this matrix will yield it's consecutive rows.

    You can use the constructor in following way:

    >>> sm = SparseMatrix([[1, 2], [3, 4]])

    to construct a matrix that looks like

    ::

        |1 2|
        |3 4|
    """
    __slots__ = ('rows_dict', 'known_column_count', 'no_cols', 'no_rows')

    def max(self) -> T:
        """
        Return maximum element.

        None elements will be ignored.
        """
        item = self[0, 0]
        for row in self:
            for value in row:
                if value is None:
                    continue
                if value > item:
                    item = value
        return item

    def min(self) -> T:
        """
        Return minimum element.

        None elements will be ignored.
        """
        item = self[0, 0]
        for row in self:
            for value in row:
                if value is None:
                    continue

                if value < item:
                    item = value
        return item

    def __init__(self, matrix_data: tp.Optional[tp.List[tp.List[T]]] = None):
        self.rows_dict = collections.defaultdict(lambda: collections.defaultdict(lambda: None))
        self.known_column_count = {}  # tp.Dict[int, int] column_no => amount
        self.no_cols = 0
        self.no_rows = 0
        if matrix_data:
            self[:, :] = matrix_data

    def get_neighbour_coordinates(self, col: int, row: int,
                                  include_diagonals: bool = True) -> tp.Iterator[
        tp.Tuple[int, int]]:
        """
        Return an iterator of coordinates to points neighbouring given point.
        :param col: column
        :param row: row
        :param include_diagonals: whether to include points having only a single point in common
        :return: an iterable of coordinates of neighbouring points. An iterator of tuple
            (col, row)
        """

        for delta_row in (-1, 0, 1):
            for delta_col in (-1, 0, 1):
                if not delta_row and not delta_col:
                    continue
                if abs(delta_col) + abs(delta_row) > 1:
                    if not include_diagonals:
                        continue
                cand_col = col + delta_col
                cand_row = row + delta_row
                if cand_col < 0 or cand_row < 0:
                    continue
                if cand_col >= self.no_cols or cand_row >= self.no_rows:
                    continue
                yield cand_col, cand_row

    def append_row(self, y: tp.Iterable[T]) -> None:
        """
        Append a row to the bottom of the matrix

        :param y: iterable with consequent columns
        """
        next_row = self.no_rows
        for col_no, z in enumerate(y):
            self[col_no, next_row] = z

    def clear(self) -> None:
        """
        Clear the contents of the sparse matrix
        """
        self.rows_dict = collections.defaultdict(lambda: collections.defaultdict(lambda: None))
        self.known_column_count = {}  # tp.Dict[int, int] column_no => amount
        self.no_cols = 0
        self.no_rows = 0

    def __eq__(self, other: 'SparseMatrix') -> bool:
        return self.rows_dict == other.rows_dict

    def __bool__(self) -> bool:
        return self.no_rows == 0

    def _sanitize_key(self, key: KeyArg) -> KeyArg:
        col, row = key
        return _cleanup_key(col, self.no_cols), _cleanup_key(row, self.no_rows)

    @classmethod
    def from_iterable(cls, y: tp.Iterable[tp.Iterable[T]]):
        """
        Construct a sparse matrix given a row-first iterable. That iterable must
        return another iterable, that will yield values for given column.

        :param y: an iterable describing the sparse matrix
        :return: a sparse matrix object
        """
        sm = SparseMatrix()
        for row_no, cols in enumerate(y):
            for col_no, value in enumerate(cols):
                sm[col_no, row_no] = value
        return sm

    @property
    def columns(self) -> int:
        """Return the amount of columns"""
        return self.no_cols

    @property
    def rows(self) -> int:
        """Return the amount of rows"""
        return self.no_rows

    def __iter__(self) -> tp.Iterator[tp.List]:
        return (self.get_row(i) for i in range(self.no_rows))

    def __len__(self) -> int:
        return self.no_rows

    def get_row(self, row_no: int) -> tp.List[T]:
        """
        Return a single row of provided number.

        The returned array has the same length as .columns

        :param row_no: row number, numbered from 0
        """
        if row_no not in self.rows_dict:  # check so as to avoid adding new entries
            output = [None] * self.no_cols
        else:
            cols = self.rows_dict[row_no]
            output = []
            for i in range(self.no_cols):
                if i in cols:
                    output.append(cols[i])
                else:
                    output.append(None)
        return output

    def shoot(self) -> 'SparseMatrix':
        """
        Insert an empty cell between current cells. So the matrix which looked like
        [[1, 2], [3, 4]] will now look like [[1, None, 2], [None, None, None], [3, None, 4]]
        """
        new_sparse = SparseMatrix()
        for row_no, row in enumerate(self):
            for col_no, value in enumerate(row):
                new_sparse[col_no * 2, row_no * 2] = value
        return new_sparse

    def _increment_column_count(self, col_no: int) -> None:
        if col_no not in self.known_column_count:
            self.known_column_count[col_no] = 1
        else:
            self.known_column_count[col_no] += 1

    def _decrement_column_count(self, col_no: int) -> None:
        if self.known_column_count[col_no] == 1:
            del self.known_column_count[col_no]
        else:
            self.known_column_count[col_no] -= 1

    def delete_row(self, row_no: int) -> None:
        """
        Delete a row with specified number

        :param row_no: number of the row to delete
        """
        cols = list(self.rows_dict[row_no].keys())  # Copy it here
        for col_no in cols:
            del self[col_no, row_no]

    def __setitem__(self, key: KeyArg,
                    value: tp.Union[tp.Iterable[T], T]) -> None:
        """
        Use this to update either a single piece of the matrix, or entire row.

        Use like:

        >>> sm[1, 2] = 5
        >>> sm[...,1] = [5]
        >>> sm[1,...] = [5]
        >>> sm[...,...] = [[5]]
        >>> sm[:,1] = [5]
        >>> sm[1,:] = [5]
        >>> sm[:,:] = [[5]]

        Note that setting an element to None is the same as deleting it.
        """
        if value is None:
            del self[key]
            return

        col, row = self._sanitize_key(key)

        if col is Ellipsis and row is Ellipsis:
            sm = SparseMatrix.from_iterable(value)
            self.rows_dict = sm.rows_dict
            self.no_rows = sm.no_rows
            self.no_cols = sm.no_cols
            self.known_column_count = sm.known_column_count
        elif col is Ellipsis:
            for col_no, val in enumerate(value):
                self[col_no, row] = val
        elif row is Ellipsis:
            for row_no, val in enumerate(value):
                self[col, row_no] = val
        else:
            if col >= self.no_cols:
                self.no_cols = col + 1
            if row >= self.no_rows:
                self.no_rows = row + 1

            if row not in self.rows_dict:
                self._increment_column_count(col)
            elif col not in self.rows_dict[row]:
                self._increment_column_count(col)
            self.rows_dict[row][col] = value

    def __getitem__(self,
                    item: KeyArg) -> tp.Union[tp.List[T], tp.List[tp.List[T]], T]:
        col, row = self._sanitize_key(item)

        if col is Ellipsis and row is Ellipsis:
            v = list(self)
        elif col is Ellipsis:
            v = [self[col_no, row] for col_no in range(self.no_cols)]
        elif row is Ellipsis:
            v = [self[col, row_no] for row_no in range(self.no_rows)]
        else:
            if row >= self.no_rows:
                raise IndexError()
            elif col >= self.no_cols:
                raise IndexError()

            if row not in self.rows_dict:  # check so as to avoid adding new entries
                v = None
            elif col not in self.rows_dict[row]:
                v = None
            else:
                v = self.rows_dict[row][col]
        return v

    @silence_excs(TypeError, returns=0)
    def _calculate_column_count(self) -> int:
        return max(self.known_column_count) + 1

    @silence_excs(TypeError, returns=0)
    def _calculate_row_count(self) -> int:
        return max(self.rows_dict) + 1

    @silence_excs(KeyError)
    def __delitem__(self, key: KeyArg) -> None:
        col, row = self._sanitize_key(key)

        if row is Ellipsis and col is Ellipsis:
            self.clear()
        elif col is Ellipsis:
            for col_no in range(self.no_cols):
                del self[col_no, row]
        elif row is Ellipsis:
            for row_no in range(self.no_rows):
                del self[col, row_no]
        else:
            # Check if the element is there
            if (row not in self.rows_dict) or (col not in self.rows_dict[row]):
                return

            del self.rows_dict[row][col]

            if not self.rows_dict[row]:  # Have we got an empty row now?
                del self.rows_dict[row]

            self._decrement_column_count(col)

            self.no_cols = self._calculate_column_count()
            self.no_rows = self._calculate_row_count()



import os
import unittest

from satella.coding import merge_dicts, update_attr_if_none


class TestMergeDicts(unittest.TestCase):
    def test_update_attr_if_none(self):
        class Object:
            def __init__(self):
                self.a = None
                self.b = 5

        a = Object()
        update_attr_if_none(a, 'a', 5)
        self.assertEqual(a.a, 5)
        update_attr_if_none(a, 'b', 4)
        update_attr_if_none(a, 'c', 8)
        self.assertEqual(a.b, 5)
        self.assertRaises(AttributeError, lambda: update_attr_if_none(
            a, 'd', 6, on_attribute_error=False))
        update_attr_if_none(a, 'b', 6, if_value_is_not_none=True)
        self.assertEqual(a.b, 6)
        self.assertEqual(a.c, 8)

    def test_advanced_merge_dicts(self):

        DA = {
            "google-fcm": {
                "http-auth": {
                    "digicort": "key=digi",
                },
                "endpoint": "https://fcm.googleapis.com/fcm/send"
            }
        }

        DB = {
            "google-fcm": {
                "http-auth": {
                    "springblade": "key=spri"
                },
                "endpoint": "https://fcm.googleapis.com/fcm/send"
            }
        }

        DC = merge_dicts(DA, DB)

        self.assertEqual(DC['google-fcm']['http-auth']['digicort'], 'key=digi')
        self.assertEqual(DC['google-fcm']['http-auth']['springblade'], 'key=spri')
        self.assertEqual(DC['google-fcm']['endpoint'], 'https://fcm.googleapis.com/fcm/send')

    def test_merge_dicts(self):

        tak = merge_dicts({'kupujemy': 'tak'}, {'kupujemy': 'nie'})
        nie = merge_dicts({'kupujemy': 'nie'}, {'kupujemy': 'tak'})

        self.assertEqual(tak['kupujemy'], 'nie')
        self.assertEqual(nie['kupujemy'], 'tak')

    def test_merge_lists(self):

        tak = merge_dicts({'kupujemy': ['tak']}, {'kupujemy': ['nie']})

        self.assertEqual(set(tak['kupujemy']), set(['tak', 'nie']))

    def test_bugA(self):
        tak = merge_dicts({}, {"a": "b"})
        self.assertEqual(tak['a'], 'b')

    def test_adds(self):
        tak = merge_dicts({'kupujemy': 'tak'}, {'xupujemy': 'nie'})

        self.assertEqual(tak['xupujemy'], 'nie')
        self.assertEqual(tak['kupujemy'], 'tak')

    def test_nest(self):
        tak = merge_dicts({'kupujemy': {"y": ['tak']}},
                          {'kupujemy': {"y": ['nie']}})

        q = tak['kupujemy']['y']

        self.assertIn('tak', q)
        self.assertIn('nie', q)



import queue
import typing as tp

from .decorators import wraps
from ..typing import ExceptionClassType, NoArgCallable, Predicate

Queue = tp.TypeVar('Queue')


def queue_get(queue_getter: tp.Union[str, tp.Callable[[object], Queue]],
              timeout: tp.Optional[float] = None,
              exception_empty: tp.Union[
                  ExceptionClassType, tp.Tuple[ExceptionClassType, ...]] = queue.Empty,
              queue_get_method: tp.Callable[[Queue, tp.Optional[float]], tp.Any] =
              lambda x, timeout: x.get(
                  timeout=timeout),
              method_to_execute_on_empty: tp.Optional[tp.Union[str, tp.Callable]] = None):
    """
    A decorator for class methods that consume from a queue.

    Timeout of None means block forever.

    First attribute of the decorator-given function must be a normal instance method
    accepting an element taken from the queue, so it must accepts two arguments - first is
    self, second is the element from the queue.

    :param queue_getter: a callable that will render us the queue, or a string, which will be
        translated to a property name
    :param timeout: a timeout to wait. If timeout happens, simple no-op will be done and None
        will be returned.
    :param exception_empty: exception (or a tuple of exceptions) that are raised on queue being
        empty.
    :param queue_get_method: a method to invoke on this queue. Accepts two arguments - the first
        is the queue, the second is the timeout. It has to follow the type signature given.
    :param method_to_execute_on_empty: a callable, or a name of the method to be executed (with no
        arguments other than self) to execute in case queue.Empty was raised. Can be a callable -
        in that case it should expect no arguments, or can be a string, which will be assumed to be
        a method name

    Use instead of:

    >>> class QueueProcessor:
    >>>     def __init__(self, queue):
    >>>         self.queue = queue
    >>>     def do(self):
    >>>         try:
    >>>             msg = self.queue.get(timeout=TIMEOUT)
    >>>         except queue.Empty:
    >>>             return

    Instead of aforementioned code, please use:

    >>> class QueueProcessor:
    >>>     def __init__(self, queue):
    >>>         self.queue = queue
    >>>     @queue_get(lambda self: self.queue, timeout=TIMEOUT)
    >>>     def do(self, msg):
    >>>         ...
    """
    if isinstance(queue_getter, str):
        def my_queue_getter(x):
            return getattr(x, queue_getter)
    else:
        my_queue_getter = queue_getter

    def outer(fun):
        @wraps(fun)
        def inner(self):
            try:
                que = my_queue_getter(self)
                item = queue_get_method(que, timeout)
                return fun(self, item)
            except exception_empty:
                if method_to_execute_on_empty is not None:
                    if callable(method_to_execute_on_empty):
                        method_to_execute_on_empty()
                    elif isinstance(method_to_execute_on_empty, str):
                        method = getattr(self, method_to_execute_on_empty)
                        method()

        return inner

    return outer


def loop_while(pred: tp.Union[Predicate, NoArgCallable[bool]] = lambda: True):
    """
    Decorator to loop the following function while predicate called on it's first argument is True.

    Use to mostly loop class methods basing off classes, like:

    >>> from satella.coding.predicates import x
    >>> class Terminable:
    >>>     terminated = False
    >>>     @loop_while(x.terminated == False)
    >>>     def run(self):
    >>>         ...

    You can also loop standard functions, like this:

    >>> a = {'terminating': False}
    >>> @loop_while(lambda: not a['terminating'])
    >>> def execute_while():
    >>>     ...

    :param pred: predicate to evaluate. Can accept either one element, in this case
        it will be fed the class instance, or accept no arguments, in which case
        it will be considered to annotate a method.

    Note that the function you decorate may only take arguments if it's a class method.
    If it's a standard method, then it should not take any arguments.
    """

    def outer(fun):
        @wraps(fun)
        def inner(*args, **kwargs):
            pred_f = pred
            if len(args):
                def pred_f():
                    return pred(args[0])
            while pred_f():
                fun(*args, **kwargs)

        return inner

    return outer



import abc
import unittest

from satella.coding import metaclass_maker, wrap_with, dont_wrap, rethrow_as, wrap_property, \
    DocsFromParent, CopyDocsFrom


class MetaA(type):
    pass


class MetaB(type):
    pass


class A(metaclass=MetaA):
    pass


class B(metaclass=MetaB):
    pass


class AB(A, B, metaclass=metaclass_maker):
    pass


class TestMetaclasses(unittest.TestCase):

    def test_copy_docs_from(self):
        class Source:
            def test(self):
                """docstring"""
        class Target(metaclass=CopyDocsFrom(Source)):
            def test(self):
                ...

        self.assertEqual(Target.test.__doc__, Source.test.__doc__)

    def test_docs_from_parent(self):
        class Father:
            """docstring"""
            def test(self):
                """my docstring"""

        class Child(Father, metaclass=DocsFromParent):
            a = 2

            def test(self):
                pass

        self.assertEqual(Child.test.__doc__, Father.test.__doc__)
        self.assertEqual(Child.__doc__, Father.__doc__)

    def test_metaclasses(self):
        AB()

    def test_wrap_all_methods_with(self):
        def double(fun):
            return lambda slf, x: fun(slf, x) * 2

        def selector(fun):
            return 'six' not in fun.__name__

        class BaseClass(abc.ABC):
            pass

        property_wrapper = wrap_property(lambda fun: rethrow_as(ValueError, KeyError)(fun))

        class Doubles(BaseClass, metaclass=wrap_with(callables=double,
                                                     properties=property_wrapper,
                                                     selector_callables=selector)):
            def return_four(self, x):
                return 2

            @property
            def key_error(self):
                raise ValueError()

            @dont_wrap
            def return_five(self, x):
                return 5

            def return_six(self, x):
                return 6

        a = Doubles()
        self.assertEqual(a.return_four(2), 4)
        self.assertEqual(a.return_five(2), 5)
        self.assertEqual(a.return_six(2), 6)
        self.assertRaises(KeyError, lambda: a.key_error)

    def test_wrap_property(self):

        def wrap_getter(fun):
            def inner(slf):
                try:
                    return fun(slf) * 2
                except TypeError:
                    return fun(slf) * 2

            return inner

        def wrap_setter(fun):
            def inner(slf, value):
                return fun(slf, value * 2)

            return inner

        def wrap_deleter(fun):
            return lambda slf: None

        wrap_prop = wrap_property(wrap_getter, wrap_setter, wrap_deleter)

        class MyClass:
            def __init__(self):
                self._x = 5

            @property
            def x(self):
                return self._x

            @x.setter
            def x(self, value):
                self._x = value

            @x.deleter
            def x(self):
                del self._x

            x = wrap_prop(x)

        mc = MyClass()
        self.assertEqual(mc.x, 10)
        mc.x = 20
        self.assertEqual(mc.x, 80)
        del mc.x
        self.assertEqual(mc.x, 80)



import enum
import typing as tp
from abc import ABCMeta, abstractmethod
import json

from satella.files import write_out_file_if_different

from satella.coding.typing import NoneType

__all__ = ['JSONEncoder', 'JSONAble', 'json_encode', 'read_json_from_file',
           'write_json_to_file', 'write_json_to_file_if_different']

Jsonable = tp.TypeVar('Jsonable', list, dict, str, int, float, None)


class JSONAble(metaclass=ABCMeta):
    __slots__ = ()

    @abstractmethod
    def to_json(self) -> Jsonable:
        """Return a JSON-able representation of this object"""


class JSONEncoder(json.JSONEncoder):
    """
    This encoder will encode everything!

    enums will be dumped to their value.
    """

    def default(self, o: tp.Any) -> Jsonable:
        if hasattr(o, 'to_json'):
            v = o.to_json()
        elif isinstance(o, (int, float, str, NoneType)):
            v = o
        elif isinstance(o, enum.Enum):
            v = o.value
        elif isinstance(o, (list, tuple)):
            v = [self.default(v) for v in o]
        elif isinstance(o, dict):
            v = {self.default(k): self.default(v) for k, v in o.items()}
        else:
            try:
                v = super().default(o)
            except TypeError:
                dct = {}
                try:
                    for k, v in o.__dict__.items():
                        dct[k] = self.default(v)
                except AttributeError:  # o has no attribute '__dict__', try with slots
                    try:
                        for slot in o.__slots__:
                            dct[slot] = self.default(getattr(o, slot))
                    except AttributeError:  # it doesn't have __slots__ either?
                        v = '<an instance of %s>' % (o.__class__.__name__,)
                v = dct
        return v


def json_encode(x: tp.Any) -> str:
    """
    Convert an object to JSON. Will properly handle subclasses of JSONAble

    :param x: object to convert
    """
    return JSONEncoder().encode(x)


def write_json_to_file(path: str, value: JSONAble, **kwargs) -> None:
    """
    Write out a JSON to a file as UTF-8 encoded plain text.

    This will use Satella's :class:`~satella.json.JSONEncoder` internally.

    :param path: path to the file
    :param value: JSON-able content
    :param kwargs: will be passed to ujson/json's dump
    """
    with open(path, 'w') as f_out:
        f_out.write(JSONEncoder().encode(value))


def write_json_to_file_if_different(path: str, value: JSONAble,
                                    encoding: str = 'utf-8', **kwargs) -> bool:
    """
    Read JSON from a file. Write out a JSON to a file if it's value is different,
    as UTF-8 encoded plain text.

    This will use Satella's :class:`~satella.json.JSONEncoder` internally.

    :param path: path to the file
    :param value: JSON-able content
    :param encoding: encoding to use while parsing the contents of the file
    :param kwargs: will be passed to ujson/json dumps
    :return: whether the write actually happened
    """
    value = JSONEncoder().encode(value)
    return write_out_file_if_different(path, value, encoding)


def read_json_from_file(path: str) -> JSONAble:
    """
    Load a JSON from a provided file, as UTF-8 encoded plain text.

    :param path: path to the file
    :return: JSON content
    :raises ValueError: the file contained an invalid JSON
    :raises OSError: the file was not readable or did not exist
    """
    try:
        import ujson
        with open(path, 'r') as f_in:
            v = ujson.load(f_in)
    except ImportError:
        with open(path, 'r') as f_in:
            try:
                v = json.load(f_in)
            except json.decoder.JSONDecodeError as e:
                raise ValueError(str(e))
    return v



import copy
import platform
import random
import sys
import threading
import time
import unittest
from concurrent.futures import ThreadPoolExecutor, Future as PythonFuture

from satella.coding.concurrent import TerminableThread, CallableGroup, Condition, MonitorList, \
    LockedStructure, AtomicNumber, Monitor, IDAllocator, call_in_separate_thread, Timer, \
    parallel_execute, run_as_future, sync_threadpool, IntervalTerminableThread, Future, \
    WrappingFuture, PeekableQueue, SequentialIssuer, CancellableCallback, ThreadCollection, \
    BogusTerminableThread, SingleStartThread, FutureCollection, MonitorSet
from satella.coding.concurrent.futures import call_in_future, ExecutorWrapper
from satella.coding.sequences import unique
from satella.exceptions import WouldWaitMore, AlreadyAllocated, Empty


class TestConcurrent(unittest.TestCase):

    def test_monitor_set(self):
        ms = MonitorSet([1, 2, 3])
        self.assertFalse(ms.insert_and_check(2))
        self.assertTrue(ms.insert_and_check(4))
        self.assertFalse(ms.insert_and_check(4))

    def test_future_collection_exception(self):
        fc = FutureCollection([PythonFuture(), PythonFuture()])
        fc.set_running_or_notify_cancel()
        fc[0].set_exception(IndexError())
        fc[1].set_exception(ValueError())
        self.assertIsInstance(fc.exception(), IndexError)

    def test_future_collection_callbacks_one(self):
        a = {'count': 0}

        def callback(fut):
            nonlocal a
            a['count'] += 1

        fc = FutureCollection([PythonFuture(), PythonFuture()])
        fc.add_done_callback(callback, True)
        fc.set_running_or_notify_cancel()
        fc.set_result(None)
        self.assertEqual(a['count'], 1)

    def test_future_collection_callbacks(self):
        a = {'count': 0}

        def callback(fut):
            nonlocal a
            a['count'] += 1

        fc = FutureCollection([PythonFuture(), PythonFuture()])
        fc.add_done_callback(callback)
        fc.set_running_or_notify_cancel()
        fc.set_result(None)
        self.assertEqual(a['count'], 2)

    def test_future_collection_cancel(self):
        fc = FutureCollection([PythonFuture(), PythonFuture()])
        self.assertTrue(fc.set_running_or_notify_cancel())

        ##
        fc = FutureCollection([PythonFuture(), PythonFuture()])
        fc.cancel()
        self.assertFalse(fc.set_running_or_notify_cancel())

        ##

        fc = FutureCollection([PythonFuture(), PythonFuture()])
        fc[0].cancel()
        self.assertTrue(fc.set_running_or_notify_cancel())

    def test_future_collection(self):
        fut1 = PythonFuture()
        fut2 = PythonFuture()
        fs = [PythonFuture(), PythonFuture()]
        fut3 = PythonFuture()

        fc = FutureCollection([fut1, fut2])
        fc.add(fut3)
        fc += fs

        self.assertIs(fc[1], fut2)

        self.assertTrue(fc.set_running_or_notify_cancel())

        @call_in_separate_thread()
        def set_future_collection():
            fc.set_result(None)

        set_future_collection()
        fc.result()

    def test_single_start_thread(self):
        dct = {'a': 0}

        class SST(SingleStartThread):
            def run(self):
                nonlocal dct
                dct['a'] += 1

        sst = SST()
        self.assertEqual(dct['a'], 0)
        sst.start()
        time.sleep(0.2)
        self.assertEqual(dct['a'], 1)
        sst.start()
        time.sleep(0.2)
        self.assertEqual(dct['a'], 1)

    def test_bogus(self):
        bt = BogusTerminableThread()
        self.assertFalse(bt.is_alive())
        self.assertRaises(RuntimeError, bt.join)
        bt.start()
        self.assertRaises(RuntimeError, bt.start)
        self.assertRaises(WouldWaitMore, lambda: bt.join(1))
        self.assertTrue(bt.is_alive())
        bt.terminate()
        self.assertRaises(RuntimeError, bt.start)
        bt.join(2)

    def test_thread_collection(self):
        dct = {}

        class Threading(threading.Thread):
            def __init__(self, a):
                super().__init__()
                self.a = a

            def run(self):
                nonlocal dct
                dct[self.a] = True
                if self.a == 6:
                    time.sleep(5)

        self.assertGreaterEqual(len(ThreadCollection.get_currently_running()), 1)
        tc = ThreadCollection.from_class(Threading, [2, 3, 4])
        tc.daemon = False
        self.assertFalse(tc.is_alive())
        self.assertEqual(len(tc), 3)
        a = Threading(5)
        tc += a
        b = [Threading(8), Threading(9)]
        tc += b
        tc += ThreadCollection([Threading(10), Threading(11)])
        self.assertFalse(tc.daemon)
        for t in tc:
            self.assertIsInstance(t, Threading)
        tc.append(Threading(6)).add(Threading(7))
        tc.start()
        self.assertRaises(WouldWaitMore, lambda: tc.join(2))
        tc.terminate().join()
        self.assertEqual(len(tc), 10)
        td = ThreadCollection([Threading(12), Threading(13)])
        te = tc + td
        self.assertEqual(len(te), 12)
        tf = te + []
        self.assertEqual(len(tf), 12)
        tg = tf + Threading(15)
        self.assertEqual(len(tg), 13)
        self.assertEqual(dct, {2: True, 3: True, 4: True, 5: True, 6: True, 7: True, 8: True,
                               9: True,
                               10: True, 11: True})

    def test_cancellable_callback(self):
        a = {'test': True}

        def y():
            nonlocal a
            a['test'] = False

        cg = CallableGroup()
        cc = CancellableCallback(y)
        cg.add(cc)
        cc.cancel()
        cg()
        self.assertTrue(a['test'])

    def test_sequential_issuer(self):
        si = SequentialIssuer()
        a = si.issue()
        b = si.issue()
        self.assertGreater(b, a)
        c = si.no_less_than(100)
        self.assertGreaterEqual(c, 100)
        d = si.issue()
        self.assertGreater(d, c)

    def test_peekable_queue_put_many(self):
        pkb = PeekableQueue()

        pkb.put_many([1, 2])

        self.assertEqual(pkb.get(), 1)
        self.assertEqual(pkb.get(), 2)

    def test_peekable_queue(self):
        pkb = PeekableQueue()

        @call_in_separate_thread()
        def put_to_queue():
            time.sleep(0.5)
            pkb.put(1)
        pkb.put(1)
        self.assertEqual(pkb.qsize(), 1)
        self.assertEqual(pkb.peek(), 1)
        self.assertEqual(pkb.qsize(), 1)
        self.assertEqual(pkb.get(), 1)
        self.assertEqual(pkb.qsize(), 0)
        put_to_queue()
        self.assertRaises(Empty, lambda: pkb.get(0.2))
        self.assertEqual(pkb.get(0.7), 1)
        put_to_queue()
        self.assertEqual(pkb.get(), 1)
        self.assertEqual(pkb.qsize(), 0)

    def test_future_on(self):
        fut = Future()
        a = {'success': False, 'exception': False}

        def on_success(arg):
            nonlocal a
            a['success'] = True

        def on_failure(exc_val):
            nonlocal a
            a['exception'] = True

        fut.on_success(on_success).on_failure(on_failure)
        fut.set_running_or_notify_cancel()
        fut.set_exception(RuntimeError())
        self.assertEqual(a, {'success': False, 'exception': True})

        fut = Future()
        a = {'success': False, 'exception': False}
        fut.on_success(on_success).on_failure(on_failure)
        fut.set_running_or_notify_cancel()
        fut.set_result(5)
        self.assertEqual(a, {'success': True, 'exception': False})

    def test_wrapped_executor(self):
        tpe = ThreadPoolExecutor(max_workers=2)
        tpe_w = ExecutorWrapper(tpe)
        a = {'test': False}

        def call_me():
            nonlocal a
            a['test'] = True

        fut = tpe_w.submit(call_me)
        self.assertIsInstance(fut, Future)
        sync_threadpool(tpe_w)
        self.assertTrue(a['test'])

    def test_wrapped_executor_nested(self):
        tpe = ThreadPoolExecutor(max_workers=2)
        tpe_w = ExecutorWrapper(ExecutorWrapper(tpe))
        a = {'test': False}

        def call_me():
            nonlocal a
            a['test'] = True

        fut = tpe_w.submit(call_me)
        self.assertIsInstance(fut, Future)
        sync_threadpool(tpe_w)
        self.assertTrue(a['test'])

    def test_call_in_future(self):
        a = {'test': False}

        def call_me(arg):
            nonlocal a
            a['test'] = arg
        tpe = ThreadPoolExecutor(max_workers=2)

        callable_ = call_in_future(tpe, call_me, True)
        self.assertFalse(a['test'])
        fut = callable_()
        self.assertIsInstance(fut, PythonFuture)
        sync_threadpool(tpe)
        self.assertTrue(a['test'])

    def test_wrapping_future(self):
        fut = PythonFuture()
        wrap = WrappingFuture(fut)
        self.assertEqual(wrap._state, fut._state)
        fut.set_running_or_notify_cancel()
        self.assertEqual(wrap._state, fut._state)
        fut.set_result(3)
        self.assertEqual(wrap.result(), 3)

    def test_wrapping_future_callback(self):
        fut = PythonFuture()
        wrap = WrappingFuture(fut)

        def transform_future(future):
            future.set_result(future.result() + 2)

        wrap.add_pre_done_callback(transform_future)
        self.assertEqual(wrap._state, fut._state)
        fut.set_running_or_notify_cancel()
        self.assertEqual(wrap._state, fut._state)
        fut.set_result(3)
        self.assertEqual(wrap.result(), 5)

    def test_future_chain(self):
        fut = Future()

        def assert_five(arg):
            self.assertEqual(arg, 5)

        fut.chain(assert_five)
        fut.set_running_or_notify_cancel()
        fut.set_result(5)

    def test_future(self):
        fut = Future()
        fut.set_running_or_notify_cancel()

        def transform_future(future):
            future.set_result(future.result() + 2)

        fut.add_pre_done_callback(transform_future)
        fut.set_result(2)
        self.assertEqual(fut.result(), 4)

    def test_future_exception(self):
        fut = Future()
        fut.set_running_or_notify_cancel()

        def transform_future(future):
            future.set_exception(ValueError())

        fut.add_pre_done_callback(transform_future)
        fut.set_exception(KeyError())
        self.assertRaises(ValueError, fut.result)

    def test_sync_threadpool(self):
        tp = ThreadPoolExecutor(max_workers=4)
        a = {'cond': 4}
        def wait():
            nonlocal a
            time.sleep(random.randint(0, 10) / 10)
            a['cond'] -= 1
        for _ in range(4):
            tp.submit(wait)
        sync_threadpool(tp)
        self.assertEqual(a['cond'], 0)

    def test_sync_threadpool_wait_max_error(self):
        tp = ThreadPoolExecutor(max_workers=1)
        tp.submit(lambda: time.sleep(3))
        self.assertRaises(WouldWaitMore, lambda: sync_threadpool(tp, 2))

    def test_sync_threadpool_wait_max_correct(self):
        tp = ThreadPoolExecutor(max_workers=1)
        tp.submit(lambda: time.sleep(0.4))
        sync_threadpool(tp, 2)

    def test_run_as_future(self):
        a = {}

        @run_as_future
        def sleep_and_set(b):
            time.sleep(2)
            a['set'] = b

        fut = sleep_and_set(True)
        self.assertNotIn('set', a)
        fut.result()
        self.assertTrue(a['set'])

        @run_as_future
        def raises():
            raise ValueError()

        fut = raises()
        self.assertRaises(ValueError, fut.result)

    def test_parallel_execute(self):
        a = {'times_called': 0}

        def return_a_future(j):
            fut = Future()
            fut.set_running_or_notify_cancel()
            if j % 2 == 0:
                fut.set_exception(Exception())
            else:
                fut.set_result(a['times_called'])
            a['times_called'] += 1
            return fut

        i = 0
        for result in parallel_execute(return_a_future, [(0,), (1,), (2,), (3,), (4,)]):
            if i % 2 == 0:
                self.assertIsInstance(result, Exception)
            else:
                self.assertEqual(result, i)
            i += 1
        self.assertEqual(5, a['times_called'])

    def test_timer_separate(self):
        a = {'test': False}

        def set_a():
            a['test'] = True

        tmr = Timer(1, set_a, spawn_separate=True)
        tmr.start()
        time.sleep(2)
        self.assertTrue(a['test'])

    def test_timer(self):
        a = {'test': False}

        def set_a():
            a['test'] = True

        tmr = Timer('1s', set_a)
        tmr.start()
        time.sleep(2)
        self.assertTrue(a['test'])

    def test_call_in_separate_thread(self):
        a = {}

        @call_in_separate_thread()
        def call_me():
            a['test'] = 2
            return 5

        fut = call_me()
        self.assertEqual(fut.result(), 5)
        self.assertEqual(a['test'], 2)

    def test_id_allocator_top_limit(self):
        id_alloc = IDAllocator(top_limit=10)
        for i in range(10):
            id_alloc.allocate_int()
        self.assertRaises(Empty, id_alloc.allocate_int)
        self.assertRaises(ValueError, lambda: id_alloc.mark_as_allocated(12))

    def test_id_allocator(self):
        id_alloc = IDAllocator()
        x = set([id_alloc.allocate_int(), id_alloc.allocate_int(), id_alloc.allocate_int()])
        y = x.pop()
        self.assertRaises(AlreadyAllocated, lambda: id_alloc.mark_as_allocated(y))
        x.add(y)
        self.assertTrue(set(unique(x)), x)
        id_alloc.mark_as_free(y)
        id_alloc.mark_as_allocated(y)

        for i in range(100):
            x.add(id_alloc.allocate_int())

        self.assertEqual(x, set(unique(x)))

        id_alloc = IDAllocator(9000)
        self.assertTrue(9000 <= id_alloc.allocate_int() <= 9009)
        self.assertTrue(9000 <= id_alloc.allocate_int() <= 9009)

    def test_atomic_number_timeout(self):
        """Test comparison while the lock is held all the time"""
        a = AtomicNumber(2)

        def process_atomic_number():
            with Monitor.acquire(a):
                time.sleep(3)

        threading.Thread(target=process_atomic_number).start()

        self.assertRaises(WouldWaitMore, lambda: a.wait_until_equal(3, timeout=1))

    def test_atomic_number_wait_equal(self):
        an = AtomicNumber()

        def process_atomic_number(a):
            time.sleep(1)
            a += 1

        threading.Thread(target=process_atomic_number, args=(an,)).start()
        self.assertRaises(WouldWaitMore, lambda: an.wait_until_equal(2, timeout=0.5))
        an.wait_until_equal(1, timeout=1)
        an.wait_until_equal(1)

        self.assertEqual(str(an), '1')
        self.assertEqual(repr(an), 'AtomicNumber(1)')

    def test_atomic_number(self):
        a = AtomicNumber(4)
        a -= 1
        self.assertIsInstance(a, AtomicNumber)
        self.assertEqual(a, 3)

        self.assertEqual(a << 1, 0b110)
        self.assertEqual(a >> 1, 1)
        self.assertEqual(1 << a, 0b1000)
        self.assertEqual(0b111000 >> a, 0b111)
        self.assertEqual(a >> 1, 1)
        self.assertIsInstance(a + 2, int)
        self.assertEqual(a | 4, 7)
        self.assertEqual(a & 2, 2)
        self.assertEqual(a ^ 3, 0)
        self.assertEqual(4 | a, 7)
        self.assertEqual(2 & a, 2)
        self.assertEqual(3 ^ a, 0)
        self.assertTrue(a > 1)
        self.assertTrue(a < 4)
        self.assertTrue(a >= 1)
        self.assertTrue(a <= 4)
        self.assertEqual(a + 1, 4)
        self.assertEqual(a - 1, 2)
        self.assertEqual(a * 1, 3)
        self.assertEqual(a / 1, 3.0)
        self.assertEqual(a // 1, 3)
        self.assertEqual(1 + a, 4)
        self.assertEqual(-1 + a, 2)
        self.assertEqual(1 * a, 3)
        self.assertEqual(1 / (a + 1), 0.25)
        self.assertEqual(1 // a, 0)
        self.assertEqual(2 ** a, 8)
        a |= 4
        self.assertEqual(a, 7)
        a &= 3
        self.assertEqual(a, 3)
        a ^= 0
        self.assertEqual(a, 3)
        a <<= 1
        self.assertEqual(a, 0b110)
        a >>= 1
        self.assertEqual(a, 3)
        a /= 1
        self.assertEqual(a, 3.0)
        a //= 1
        self.assertEqual(a, 3)
        self.assertEqual(a ** 2, 9)
        a **= 2
        self.assertEqual(a, 9)
        self.assertEqual(+a, 9)
        self.assertTrue(bool(a))
        self.assertEqual(int(a), 9)
        self.assertEqual(float(a), 9.0)
        self.assertEqual(-a, -9)
        self.assertEqual(abs(a), 9)

    def test_locked_structure(self):
        class MyLockedStructure:
            def __call__(self, *args, **kwargs):
                return 5

            def __init__(self, a, b, c):
                self.a = a
                self.b = b
                self.c = c

        mls = MyLockedStructure(1, 2, 3)
        ls = LockedStructure(mls)

        ls.b = 5
        with ls:
            self.assertEqual(mls.b, 5)
            self.assertEqual(mls.a, 1)
            self.assertEqual(mls(), 5)

    def test_copy_monitor_list(self):
        ml = MonitorList([1, 2, 3])
        ml2 = copy.copy(ml)
        self.assertEqual(ml2, ml)

    def test_terminate_on_exception_in_prepare(self):
        slf = self

        class MyThread(TerminableThread):
            def __init__(self):
                super().__init__(terminate_on=(ValueError, KeyError))
                self.a = 0

            def prepare(self) -> None:
                raise ValueError()

            def loop(self) -> None:
                slf.fail('Cannot invoke loop on a terminated thread!')

        mt = MyThread()
        mt.start().join()

    def test_terminate_on_exception(self):
        class MyThread(TerminableThread):
            def __init__(self):
                super().__init__(terminate_on=(ValueError, KeyError))
                self.a = 0

            def loop(self) -> None:
                self.a += 1
                if self.a == 4:
                    raise ValueError()

        mt = MyThread()
        mt.start().join()
        mt.terminate()

    @unittest.skipUnless(sys.implementation.name == 'cpython', 'Does not work on PyPy :(')
    def test_condition(self):
        dct = {'a': False}
        slf = self
        cond = Condition()
        interlock_cond = Condition()

        class MyThread(TerminableThread):
            def run(self) -> None:
                self.safe_sleep(-2)
                self.safe_sleep(0.5)
                interlock_cond.notify()
                slf.assertRaises(WouldWaitMore, lambda: cond.wait(timeout=0.5))
                cond.wait(timeout=0.5, dont_raise=True)
                cond.wait()
                dct['a'] = True

        MyThread().start()
        interlock_cond.wait()
        self.assertFalse(dct['a'])
        time.sleep(2)
        cond.notify()
        time.sleep(0.1)
        self.assertTrue(dct['a'])

    def test_terminablethread_condition(self):
        a = {'dct': False}
        condition = Condition()
        slf = self

        class MyThread(TerminableThread):
            def __init__(self):
                super().__init__()

            def run(self) -> None:
                nonlocal a, slf, condition
                self.safe_wait_condition(condition, 5)
                a['dct'] = True
                slf.assertRaises(WouldWaitMore, lambda: self.safe_wait_condition(condition, 3))

        t = MyThread().start()
        time.sleep(0.2)
        self.assertTrue(t.is_alive())
        self.assertFalse(a['dct'])
        condition.notify()
        time.sleep(0.1)
        self.assertTrue(a['dct'])

    def test_terminate_on(self):
        dct = {'a': False}

        class MyThread(TerminableThread):
            def __init__(self):
                super().__init__(terminate_on=ValueError)

            def loop(self) -> None:
                nonlocal dct
                if dct['a']:
                    raise ValueError()

        t = MyThread().start()
        self.assertTrue(t.is_alive())
        time.sleep(1)
        self.assertTrue(t.is_alive())
        dct['a'] = True
        time.sleep(1)
        self.assertTrue(t.terminating)
        self.assertFalse(t.is_alive())

    def test_cg_proforma(self):
        cg = CallableGroup()
        a = {}
        cg.add(lambda: a.__setitem__('test', 'value'))
        cg()
        self.assertEqual(a['test'], 'value')

    def test_terminable_thread(self):
        class MyTerminableThread(TerminableThread):
            def prepare(self):
                self.a = 5

            def loop(self):
                self.a += 1
                time.sleep(0.5)

        mtt = MyTerminableThread()
        mtt.start()
        mtt.terminate().join()

    def test_interval_terminable_thread(self):
        class MyTerminableThread(IntervalTerminableThread):
            def __init__(self):
                super().__init__(1)
                self.a = 0
                self.overrun = False

            def prepare(self) -> None:
                self.a = 1

            def on_overrun(self, time_taken: float) -> None:
                self.overrun = True

            def loop(self) -> None:
                if self.a == 3:
                    time.sleep(3)
                self.a += 1

        mtt = MyTerminableThread()
        mtt.start()
        time.sleep(0.3)
        self.assertIn(mtt.a, (1, 2))
        self.assertFalse(mtt.overrun)
        time.sleep(1.2)
        self.assertFalse(mtt.overrun)
        time.sleep(4)
        self.assertTrue(mtt.overrun)
        mtt.terminate().join()

    def test_interval_terminable_thread_terminates(self):
        class MyTerminableThread(IntervalTerminableThread):
            def __init__(self, a):
                super().__init__(1, terminate_on=ValueError)
                self.a = a
                self.overrun = False

            def prepare(self) -> None:
                if self.a:
                    raise ValueError()

            def loop(self) -> None:
                if not self.a:
                    raise ValueError()

        mtt_a = MyTerminableThread(True)
        mtt_b = MyTerminableThread(False)
        mtt_a.start()
        mtt_b.start()
        mtt_a.terminate().join()
        mtt_b.terminate().join()

    @unittest.skipIf(platform.python_implementation() != 'PyPy', 'this requires PyPy')
    def test_terminable_thread_force_notimplementederror(self):
        class MyTerminableThread(TerminableThread):

            def run(self):
                a = 0
                while not self.terminating:
                    time.sleep(0.5)
                    a += 1

        mtt = MyTerminableThread(daemon=True)
        mtt.start()
        self.assertRaises(NotImplementedError, lambda: mtt.terminate(force=True))
        mtt.terminate().join()

    @unittest.skipIf(platform.python_implementation() == 'PyPy', 'force=True doesn''t work on PyPy')
    def test_terminable_thread_force(self):
        class MyTerminableThread(TerminableThread):
            def run(self):
                a = 0
                while True:
                    time.sleep(0.5)
                    a += 1

        mtt = MyTerminableThread(daemon=True)
        mtt.start()
        mtt.terminate(force=True)
        mtt.join(3)
        self.assertFalse(mtt.is_alive())

    def test_callable_group_some_raise(self):
        cg = CallableGroup(gather=True)
        cg.add(lambda: dupa)
        cg.add(lambda: 5)
        self.assertRaises(NameError, lambda: cg())

        cg = CallableGroup(gather=True, swallow_exceptions=True)
        cg.add(lambda: dupa)
        cg.add(lambda: 5)
        self.assertEqual(cg()[1], 5)
        self.assertIsInstance(cg()[0], NameError)

    def test_callable_group(self):
        a = {
            'a': False,
            'b': False
        }

        def op_f(what):
            return lambda: a.__setitem__(what, True)

        cg = CallableGroup()

        cg.add(op_f('a'))
        cg.add(op_f('b'))

        cg()

        self.assertTrue(all(a.values()))



import unittest

from satella.coding import assert_equal, Inequal


class TestDeepCompare(unittest.TestCase):
    def assertInequal(self, a, b):
        self.assertRaises(Inequal, lambda: assert_equal(a, b))

    def test_compare(self):
        assert_equal(2, 2)
        assert_equal(2.0, 2.0)
        assert_equal(None, None)
        assert_equal('test', 'test')
        assert_equal(b'test', b'test')
        self.assertInequal(3, 2)
        self.assertInequal([], [6])
        self.assertInequal({1: 2}, {3: 4})
        self.assertInequal({1: 2}, {1: 4})
        self.assertInequal({1: 2}, {3: 4, 5: 6})
        assert_equal([1], [1])
        assert_equal(set([1, 2]), set([2, 1]))

        class Object:
            def __init__(self, a):
                self.a = a

            def __eq__(self, other):
                return self.a == other.a

        assert_equal(Object(3), Object(3))
        self.assertInequal(Object(3), Object(4))



import enum


class InequalityReason(enum.IntEnum):
    NOT_EQUAL = 0           #: direct eq yielded not equal
    LENGTH_MISMATCH = 1     #: length didn't match
    KEY_NOT_FOUND = 2       #: key given as obj1 was not found


class Inequal(Exception):
    """
    An exception raised by :meth:`~satella.coding.deep_compare` if two objects don't match

    :ivar obj1: first object that was not equal, or key name
    :ivar obj2: second object that was not equal, or None
    :ivar reason: (:class:`~satella.coding.InequalityReason`) reason for inequality
    """
    def __init__(self, obj1, obj2, reason: InequalityReason):
        self.obj1 = obj1
        self.obj2 = obj2
        self.reason = reason


def assert_equal(a, b):
    """
    Assert that two values are equal. If not, an :class:`satella.coding.Inequality` exception
    will be thrown.

    Objects are tried to compare using it's :code:`__eq__`.

    :param a: first value to compare
    :param b: second value to compare
    :raises Inequal: objects were not equal
    """
    if isinstance(a, (int, float, str, bytes, bytearray, type(None))):
        if a != b:
            raise Inequal(a, b, InequalityReason.NOT_EQUAL)
        return
    elif isinstance(a, (set, frozenset)):
        if len(a) != len(b):
            raise Inequal(a, b, InequalityReason.LENGTH_MISMATCH)
        if a != b:
            raise Inequal(a, b, InequalityReason.LENGTH_MISMATCH)
    elif isinstance(a, (list, tuple)):
        if len(a) != len(b):
            raise Inequal(a, b, InequalityReason.LENGTH_MISMATCH)
        for c, d in zip(a, b):
            assert_equal(c, d)
    elif isinstance(a, dict):
        if len(a) != len(b):
            raise Inequal(a, b, InequalityReason.LENGTH_MISMATCH)
        for key in a.keys():
            if key not in b:
                raise Inequal(key, None, InequalityReason.KEY_NOT_FOUND)
            assert_equal(a[key], b[key])
    else:
        try:
            if a == b:
                return
            raise Inequal(a, b, InequalityReason.NOT_EQUAL)
        except TypeError:
            pass




import typing as tp


def is_subset(subset: tp.Dict, superset: tp.Dict) -> bool:
    """
    Does superset contain all keys of subset, and are their values equal?

    :param subset: the set that contains all the keys
    :param superset: the set that is to contain all the keys in subset, and their values
        have to be equal
    :return: does the condition hold?
    """
    for k, v in subset.items():
        if k not in superset:
            return False
        if v != superset[k]:
            return False
    return True



import queue
import typing as tp

from satella.coding.concurrent.monitor import Monitor
from satella.coding.recast_exceptions import silence_excs
from satella.coding.typing import T


class Subqueue(tp.Generic[T]):
    """
    Or a named queue is a collection of thread safe queues identified by name
    """

    def __init__(self):
        self.subqueues = {}  # type: tp.Dict[str, queue.Queue]
        self.subqueue_lock = Monitor()

    def assert_queue(self, queue_name: str) -> None:
        """
        Assure that we have a queue with a particular name in the dictionary
        """
        if queue_name not in self.subqueues:
            with self.subqueue_lock:
                if queue_name not in self.subqueues:  # double check for locking
                    self.subqueues[queue_name] = queue.Queue()

    def put(self, queue_name: str, obj: T) -> None:
        """
        Same semantics as queue.Queue.put
        """
        self.assert_queue(queue_name)
        self.subqueues[queue_name].put(obj)

    def get(self, queue_name: str, block=True, timeout=None) -> T:
        """
        Same semantics at queue.Queue.get
        """
        self.assert_queue(queue_name)
        return self.subqueues[queue_name].get(block, timeout)

    def qsize(self) -> int:
        """Calculate the total of entries"""
        return sum(que.qsize() for que in self.subqueues.values())

    def get_any(self) -> tp.Tuple[str, T]:
        """
        Return any message with name of the queue.

        This assumes block=False

        :return: a tuple of (queue_name, object)
        :raises: queue.Empty()
        """
        for queue_name, q in self.subqueues.items():
            with silence_excs(queue.Empty):
                return queue_name, q.get(block=False)
        else:
            raise queue.Empty('No messages in any of the queues')
